"""
This module contains a few small sphinx extensions.
They are mainly used to help with the generation
of BPS's own documentation, but some other projects
use them as well, so they are kept here.
"""
import re
import os.path

__version__ = "1.1"

def get_theme_dir():
    "return path to directory containing sphinx themes in this package"
    return os.path.abspath(os.path.join(__file__,os.path.pardir, "themes"))

def get_version(release):
    "derive short version string from longer release"
    return re.match("(\d+\.\d+)", release).group(1)



"""
===================================================================================
:mod:`cloud_sptheme.ext.autodoc_sections` - support for ReST sections in docstrings
===================================================================================

This Sphinx extension should be used in conjunction with the Autodoc extension.
This extension allows docstrings (module, class, function, etc) to include
ReST-style headers; which normally cause problems if integrated
into Sphinx documentation via Autodoc.

This extension detects all single-line (not double-lined) headers,
and turned them into a series of definition lists, with the header
name as the term, and the header's content as the definition.

In order to allow additional styling to be added, each definition entry
has two html css styles added: ``nested-section``; and :samp:`nested-section-{level}`,
which allows styling of different indentation levels for each section.
The :doc:`/cloud_theme` natively provides css styling for these classes.

.. warning::

    This extension is currently somewhat fragile,
    it works reliably for the common cases,
    but may not behave properly given unexpected input.
"""
import re
import logging; log = logging.getLogger(__name__)

def indent_sections(lines, reference_prefix=''):
    "replaces any section headers with indented paragraphs"
    end = len(lines)-1
    out = []

    sections = []
    indent_char = ' ' * 4
    indent_level = 0
    SCHARS = '#*=-^"'
    section_chars = [""] #map of section char -> level (built up as scan progresses)
    def get_level(c):
        sc = section_chars[0]
        if c not in sc:
            sc = section_chars[0] = (sc+c)
        return sc.index(c)
    #FIXME: this doesn't detect double-barred sections
    def detect_section(idx):
        if idx == end:
            return None
        line = lines[idx].rstrip()
        if not line or line.lstrip() != line:
            return None
        next = lines[idx+1].rstrip()
        if next.lstrip() != next:
            return None
        for c in SCHARS:
            if next.startswith(c * len(line)):
                return c
        return None
    idx = 0
    lss = False #set to true last non-empty line was a section heading
    while idx <= end:
        line = lines[idx].rstrip()
        if not line:
            if not lss:
                out.append("")
            idx += 1
            continue
        new_char = detect_section(idx)
        if new_char:
            new_level = get_level(new_char)
            while sections and sections[-1] > new_level:
                sections.pop()
            if not sections or sections[-1] < new_level:
                sections.append(new_level)
            name = line.lower().strip().replace(" ", "-").replace("--", "-")
            indent = indent_char * (indent_level-1)
            out.extend([
                indent + ".. _%s:" % (reference_prefix + name),
                indent + ".. rst-class:: nested-section nested-section-%d" % (new_level+1,),
                "",
                indent + "%s\n" % line.rstrip(),
                ])
            idx += 2 #skip section header
            indent_level = max(0, len(sections))
            lss = True
            continue
        lss = False
        indent = indent_char * indent_level
        out.append(indent + line)
        idx += 1
    return out

def _remove_oneline(name, lines):
    #remove one-line description from top of module, if present,
    #cause we don't want it being duplicated (should already be listed in module's header)
    _title_re = re.compile(r"""
        ^ \s*
        ( {0} \s* -- \s* )?
        [a-z0-9 _."']*
        $
    """.format(re.escape(name)), re.X|re.I)
    if len(lines) > 1 and _title_re.match(lines[0]) and lines[1].strip() == '':
        del lines[:2]

def mangle_docstrings(app, what, name, obj, options, lines):
    if what == 'module':
        _remove_oneline(name, lines)
    elif what in ('class', 'exception', 'function', 'method'):
        name = "%s.%s" % (obj.__module__, obj.__name__)
        name = name.replace(".", "-").lower()
        lines[:] = indent_sections(lines, reference_prefix=name + "-")
    elif what in ('attribute',):
        pass
    else:
        #FIXME: handle other cases
        raise NotImplementedError, "unknown node: %r %r" % (what, obj)

def setup(app):
    app.connect('autodoc-process-docstring', mangle_docstrings)



"""
=================================================================
:mod:`cloud_sptheme.index_styling` - adds css styling to genindex
=================================================================

This sphinx extension intercepts & modifies the general index data
before it is rendered to html; and does the following:

All index entries are wrapped in ``<span>`` elements,
where each element has the format :samp:`<span class="category {type}">`.
:samp:`{type}` in turn is one of ``attribute``, ``method``, ``class``, ``function``, ``module``.
The name of the entry itself is wrapped in an addtional ``<span class="subject">`` element.
Entries which don't fit into one of the above categories are not modified.

The purpose of this class is solely to allow themes (such as the :doc:`/cloud_theme`)
to provide addtional per-type styling for index entries.
"""
import logging; log = logging.getLogger(__name__)
import re
from jinja2 import Markup as literal, escape

prefix = r"^(?P<name>.*)\("
suffix = r"\)$"
_attr_re = re.compile(prefix + r"(?P<left>)(?P<sub>.*)(?P<right> attribute)" + suffix)
_meth_re = re.compile(prefix + r"(?P<left>)(?P<sub>.*)(?P<right> method)" + suffix)
_fc_re = re.compile(prefix + r"(?P<left>class in |in module )(?P<sub>.*)(?P<right>)" + suffix)
_mod_re = re.compile(prefix + r"module" + suffix)

def format_index_name(name):
    while True:
        m = _attr_re.match(name)
        if m:
            name, left, sub, right = m.group("name","left", "sub", "right")
            type = "attribute"
            break
        m = _meth_re.match(name)
        if m:
            name, left, sub, right = m.group("name","left", "sub", "right")
            type = "method"
            break
        m = _fc_re.match(name)
        if m:
            name, left, sub, right = m.group("name","left", "sub", "right")
            if left.startswith("class"):
                type = "class"
            else:
                type = "function"
            break
        m = _mod_re.match(name)
        if m:
            name = m.group("name")
            left = "module"
            sub = right = ''
            type = "module"
            break
        return name
    if sub:
        sub = literal('<span class="subject">') + escape(sub) + literal("</span>")
    cat = left + sub + right
    return escape(name) + literal('<span class="category ' + type + '">(') + escape(cat) + literal(")</span>")

def mangle_index(app, pagename, templatename, ctx, event_arg):
    if pagename != "genindex":
        return
    fmt = format_index_name
    for key, entries in ctx['genindexentries']:
        for idx, entry in enumerate(entries):
            name, (links, subitems) = entry
            entries[idx] = fmt(name), (links, subitems)
            for idx, entry in enumerate(subitems):
                name, links = entry
                subitems[idx] = fmt(name), links

def setup(app):
    app.connect('html-page-context', mangle_index)
